import math
import torch
import numpy as np
import matplotlib.pyplot as plt

def freedom_search(X, k, l):
    """ Computes the number of freedoms for the connected component at
    position (k, l) on board X.

    X: A m x m matrix where -1 entries indicate stones of the opposing player,
       0 entries indicate free fields, and +1 entries indicate own stones.
    k: The row of the position for which the number of freedoms shall be
       computed.
    l: The column of the position for which the number of freedoms shall be
       computed.

    Returns:
    freedoms: The number of freedoms for the connected component at position
              (k, l).
    members:  The member fields of the connected component at position (k, l).
    """
    color = X[k, l]
    if torch.abs(color) < 0.5:
        raise ValueError('Tried to compute freedoms for an empty field')
    visited = torch.zeros(X.shape[0], X.shape[1], dtype=torch.uint8)
    # start a depth first search to identify the connected component
    # this field is part of
    stk = [(k, l)]
    members = []
    freedoms = 0
    while stk:
        i, j = stk.pop()
        if visited[i, j]:
            continue
        # mark the field as visited
        visited[i, j] = 1
        # if this is a free field, add a freedom to the connected component
        # and continue
        if torch.abs(X[i, j]) < 0.5:
            freedoms += 1
            continue
        # otherwise, continue the search at adjacent fields
        if i > 0:
            if X[i-1, j] * color < -0.5:
                pass
            else:
                stk.append((i-1, j))
        if i < X.shape[0]-1:
            if X[i+1, j] * color < -0.5:
                pass
            else:
                stk.append((i+1, j))
        if j > 0:
            if X[i, j-1] * color < -0.5:
                pass
            else:
                stk.append((i, j-1))
        if j < X.shape[1]-1:
            if X[i, j+1] * color < -0.5:
                pass
            else:
                stk.append((i, j+1))
        # mark the current field as part of the connected component
        members.append((i, j))
    return freedoms, members

def apply_move(X, i, j):
    """ Applies the move X[i, j] = 1 to the given board X.

    Args:
    X: A m x m matrix where -1 entries indicate stones of the opposing player,
       0 entries indicate free fields, and +1 entries indicate own stones.
    i: The row of the new stone.
    j: The column of the new stones.

    Returns:
    X: The new board state.
    """

    # set the field on the board
    X[i, j] = 1.
    # check if any enemy stones are removed from the field
    removed = []
    if i > 0:
        if X[i-1, j] < -0.5:
            freedoms, members = freedom_search(X, i-1, j)
            # stones are removed if we block the last freedom
            # with our move
            if freedoms < 0.5:
                # remove all stones in the connected component
                removed += members
    if i < X.shape[0] - 1:
        if X[i+1, j] < -0.5:
            freedoms, members = freedom_search(X, i+1, j)
            # stones are removed if we block the last freedom
            # with our move
            if freedoms < 0.5:
                # remove all stones in the connected component
                removed += members
    if j > 0:
        if X[i, j-1] < -0.5:
            freedoms, members = freedom_search(X, i, j-1)
            # stones are removed if we block the last freedom
            # with our move
            if freedoms < 0.5:
                # remove all stones in the connected component
                removed += members
    if j < X.shape[1] - 1:
        if X[i, j+1] < -0.5:
            freedoms, members = freedom_search(X, i, j+1)
            # stones are removed if we block the last freedom
            # with our move
            if freedoms < 0.5:
                # remove all stones in the connected component
                removed += members

    for (i, j) in removed:
        X[i, j] = 0.

    return X

def compute_freedoms(X):
    """ Computes the freedoms of all stones on the board, i.e. the number of
    free fields adjacent to the connected component a stone is part of. The
    freedoms of free fields are computed as the number of freedoms that a stone
    on this field would have.

    Args:
    X: A m x m matrix where -1 entries indicate stones of the opposing player,
       0 entries indicate free fields, and +1 entries indicate own stones.

    Returns:
    Y: A m x m matrix where each field lists the number of freedoms of the
       connected component this field is part of.
    """
    # do a first run over the board where we compute the freedoms for both
    # players
    Y = torch.zeros(X.shape[0], X.shape[1])
    visited = torch.zeros(X.shape[0], X.shape[1], dtype=torch.uint8)
    for k in range(X.shape[0]):
        for l in range(X.shape[1]):
            # skip this field if it was already visited or if it does not
            # belong to any player
            if torch.abs(X[k, l]) < 0.5 or visited[k, l]:
                continue
            # start a depth first search to identify the connected component
            # this field is part of
            freedoms, members = freedom_search(X, k, l)
            # once the depth-first-search is finished, write the same number
            # of freedoms in all cells of the connected component
            for (i, j) in members:
                Y[i, j] = freedoms
                visited[i, j] = 1
    # fo a second run over the board and compute the freedoms for all fields
    # that are currently free under the assumption that the player places their
    # stone there
    for k in range(X.shape[0]):
        for l in range(X.shape[1]):
            if torch.abs(X[k, l]) > 0.5:
                continue
            # apply the move to a copy of the board
            Xnext = apply_move(X.clone().detach(), k, l)
            # start a depth first search to identify the connected component
            # this field is part of
            freedoms, members = freedom_search(Xnext, k, l)
            Y[k, l] = freedoms
    return Y

# a simulator function which runs a game of Go on a 5 x 5 board between two policies
def simulate_game(pol_black, pol_white = None, board_size = 5, max_moves = None):
    """ Simulates a game of go between the two given policies on a board of the
    given size. Black starts.

    Args:
    pol_black: A policy, i.e. a function which receives the current board
               state as input and returns a matrix of equal size where the
               entries are interpreted as softmax weights. For example, for
               a board of size 2 x 2 we expect as response a 2 x 2 matrix W
               with entries [[w11, w12], [w21, w22]] and the move (i, j) will
               then be selected with probability exp(wij) /
               ( exp(w11) + exp(w12) + exp(w21) + exp(w22)). This function will
               automatically exclude invalid moves.
    pol_white: See pol_black. If this is not given, a simple policy is used
               that maximizes the freedoms of own stones.
    board_size: The size of the Go board. 5 per default.
    max_moves: The maximum number of moves until the simulation ends.
               board_size x board_size per default.

    Returns:
    trace:     A sequence of board states, i.e. a list of
               board_size x board_size matrices where -1 entries indicate white
               stones, 0 entries indicate empty fields, and +1 entries indicate
               black stones.
    actions:   A sequence of actions, i.e. a list of tuples (i, j) such that at
               time t a stone was placed at (i, j). If a player passed, the entry
               is None.
    possible_actions: A sequence of possible actions, i.e. a list of boolean
               board_size x board_size matrices where a True value indicates that
               the player could have placed a stone at this position at this time.
    reward:    +1 if black won and -1 if white won. Note that white wins if the
               number of stones on the board is equal.
    """
    if max_moves is None:
        max_moves = board_size * board_size
    else:
        max_moves = math.ceil(max_moves / 2)

    board = torch.zeros(board_size, board_size)
    trace = [board.clone().detach()]
    actions = []
    possible_actions = []
    for t in range(max_moves):
        # black begins
        # compute possible moves
        # check the freedoms of each field
        Y = compute_freedoms(board)
        valid_fields = np.logical_and(Y.detach().numpy() > 0.5, np.abs(board.detach().numpy()) < 0.5)
        # apply black policy
        W = pol_black(board)
        black_passed = W is None or not np.any(valid_fields)
        if not black_passed:
            # compute softmax probabilities from policy using numpy
            P = W.detach().numpy()
            # subtract maximum for numerical reasons
            P -= np.max(P[valid_fields])
            # take exponential
            P = np.exp(P)
            # set impossible moves to zero
            P[np.logical_not(valid_fields)] = 0.
            # check that at least one valid move exists
            if np.sum(P) > 0.5:
                # normalize
                P /= np.sum(P)
                # sample
                k = np.random.choice(board_size * board_size, 1, p = P.ravel())[0]
                i = int(k / board_size)
                j = k % board_size
                board = apply_move(board, i, j)
            else:
                black_passed = True
        trace.append(board)
        possible_actions.append(valid_fields)
        if black_passed:
            actions.append(None)
        else:
            actions.append((i, j))
        # flip the board
        board = -board
        # compute possible moves
        # check the freedoms of each field
        Y = compute_freedoms(board)
        valid_fields = np.logical_and(Y.detach().numpy() > 0.5, np.abs(board.detach().numpy()) < 0.5)
        # apply white policy
        if pol_white is None:
            # if no policy is given, use a default policy where we use the
            # number of freedoms as basic utility and add 1 if we can place
            # an enemy stone in Atari and 10 if we can beat an enemy stone
            W = Y
            for i in range(board.shape[0]):
                for j in range(board.shape[1]):
                    if not valid_fields[i, j]:
                        continue
                    # check neighbors
                    if i > 0 and board[i-1, j] < -0.5:
                        if Y[i-1, j] < 1.5:
                            W[i, j] += 10
                        elif Y[i-1, j] < 2.5:
                            W[i, j] += 1
                    if i < board.shape[0]-1 and board[i+1, j] < -0.5:
                        if Y[i+1, j] < 1.5:
                            W[i, j] += 10
                        elif Y[i+1, j] < 2.5:
                            W[i, j] += 1
                    if j > 0 and board[i, j-1] < -0.5:
                        if Y[i, j-1] < 1.5:
                            W[i, j] += 10
                        elif Y[i, j-1] < 2.5:
                            W[i, j] += 1
                    if j < board.shape[1] -1 and board[i, j+1] < -0.5:
                        if Y[i, j+1] < 1.5:
                            W[i, j] += 10
                        elif Y[i, j+1] < 2.5:
                            W[i, j] += 1
            # if there is no good move, let the policy pass
            if np.sum(valid_fields) < 0.5 or np.max(W.detach().numpy()[valid_fields]) < 1.5:
                W = None
        else:
            W = pol_white(board)
        white_passed = W is None or not np.any(valid_fields)
        if not white_passed:
            # compute softmax probabilities from policy using numpy
            P = W.detach().numpy()
            # subtract maximum for numerical reasons
            P -= np.max(P[valid_fields])
            # take exponential
            P = np.exp(P)
            # set impossible moves to zero
            P[np.logical_not(valid_fields)] = 0.
            # check that at least one valid move exists
            if np.sum(P) > 0.5:
                # normalize
                P /= np.sum(P)
                # sample
                k = np.random.choice(board_size * board_size, 1, p = P.ravel())[0]
                i = int(k / board_size)
                j = k % board_size
                board = apply_move(board, i, j)
            else:
                white_passed = True
        trace.append(-board)
        possible_actions.append(valid_fields)
        if white_passed:
            actions.append(None)
        else:
            actions.append((i, j))
        # flip the board
        board = -board
        # if both players passed, end the game
        if black_passed and white_passed:
            break
    # check who won via simplified chinese rules, i.e.
    # check who has the most stones on the board
    black_stones = (board > 0.5).sum()
    white_stones = (board < -0.5).sum()
    if black_stones - white_stones > 0.5:
        reward = 1.
    else:
        reward = -1.
    return trace, actions, possible_actions, reward

def visualize_board(X, axis = None):
    """ Visualizes the given board state.

    Args:
    X: A m x m matrix where -1 entries indicate white stones,
       0 entries indicate free fields, and +1 entries indicate black stones.
    axis: The axis handle to which the visualization should be plotted.
       Creates a new axis per default.

    Returns:
    axis: The axis handle for the visualization.
    """
    if axis is None:
        _, axis = plt.subplots()

    m, n = X.shape
    # plot all lines first
    for i in range(m):
        axis.plot([0, n-1], [-i, -i], 'k-')
    for j in range(n):
        axis.plot([j, j], [0, -m+1], 'k-')
    # plot the stones
    for i in range(m):
        for j in range(n):
            if X[i, j] < -0.5:
                axis.add_artist(plt.Circle((j, -i), 0.35, edgecolor = 'black', facecolor = 'white', zorder = 10))
            elif X[i, j] > 0.5:
                axis.add_artist(plt.Circle((j, -i), 0.35, color = 'black'))
    axis.set_xlim((-0.5, n-0.5))
    axis.set_ylim((-m+0.5,0.5))
    axis.set_aspect('equal', 'box')
    axis.set_axis_off()
    return axis

def visualize_game(trace):
    """ Visualizes a game of go with the given state of board
    states.

    trace: A list of board states. Each state will be visualized
           individually.

    Returns:
    fig: The figure handle for the visualization.
    axs: The axes handles for the visualization.
    """
    T = len(trace)
    m = 4
    n = math.ceil(T / m)

    fig, axs = plt.subplots(n, m, figsize=(10, 3 * n))

    for t in range(T):
        visualize_board(trace[t], axs[int(t / m), t % m])
        axs[int(t / m), t % m].set_title('%d' % t)
    if T % m != 0:
        for t in range(T % m, m):
            axs[n-1, t].set_axis_off()
    fig.tight_layout()
    return fig, axs

# create one reference game for visualization purposes
example_moves = [(1, 1), (1, 0), (0, 1), (2, 1), (1, 2), (2, 0), (2, 2), None, (0, 0), (2, 1), (2, 0)]
X = torch.zeros(3, 3)
example_game  = [X.clone().detach()]
for t in range(len(example_moves)):
    if example_moves[t] is not None:
        X = apply_move(X, *example_moves[t])
    X = -X
    if t % 2 == 0:
        example_game.append(-X)
    else:
        example_game.append(X.clone().detach())

# test the game simulation in a deterministic case
t = -1
def deterministic_policy(board):
    global t
    # increment t
    t += 1
    # pass if so told
    if t >= len(example_moves) or example_moves[t] is None:
        return None
    # otherwise generate a weight matrix with zeros and value 30
    # for the desired move. After softmax application, this should
    # ensure that the desired action is indeed taken.
    W = torch.zeros(board.shape)
    i, j = example_moves[t]
    W[i, j] = 30.
    return W

trace, actions, possible_actions, reward = simulate_game(deterministic_policy, deterministic_policy, 3)
# check result
for t in range(len(example_moves)):
    if example_moves[t] != actions[t]:
        raise ValueError('expected action %s in step %d but got action %s' % (example_moves[t], t, actions[t]))
